"""Layered graph layouts with thin edges, as a byproduct of the Sugiyama method"""

from typing import Any, Dict, Optional, Tuple

import networkx as nx

from . import processing


def get_layered_graph_layout(
    dig: nx.DiGraph, layout_config: Optional[processing.LayoutConfig] = None
) -> Dict[Any, Tuple[float, float]]:
    """Get layered graph layout for a directed graph"""
    dig = dig.copy()
    for edge in dig.edges:
        dig.edges[edge]["value"] = 1

    processing.check_digraph_before_processing(dig)
    processing.process_directed_graph(dig, layout_config=layout_config)

    pos = {node: (dig.nodes[node]["layer"], dig.nodes[node]["y"]) for node in dig.nodes}
    return pos



"""Example data to test and showcase the package"""

import copy
from typing import Dict

import networkx as nx
import pandas as pd


def balanced_tree(branching_factor: int = 2, height: int = 3):
    """Balanced tree"""
    tree = nx.balanced_tree(branching_factor, height, create_using=nx.DiGraph)
    dig = get_digraph_with_edge_values_from_tree(tree)
    return dig


def siamese_balanced_tree(branching_factor=2, height=3):
    """Balanced tree with mirrored copy"""
    tree = nx.balanced_tree(branching_factor, height, create_using=nx.DiGraph)
    dig = get_digraph_with_edge_values_from_tree(tree)
    dig = get_mirrored_graph(dig)
    return dig


def balanced_tree_with_cross_edge(branching_factor=2, height=3):
    """Balanced tree with an additional cross edge"""
    tree = nx.balanced_tree(branching_factor, height, create_using=nx.DiGraph)
    dig = get_digraph_with_edge_values_from_tree(tree)
    cross_edge, edge_value = (2, 4), 20
    add_cross_edge(dig, cross_edge, edge_value)
    return dig


example_fun_list_no_crossing = [
    balanced_tree,
    balanced_tree_with_cross_edge,
    siamese_balanced_tree,
]

example_fun_list = list(example_fun_list_no_crossing)


def get_example_with_simple_cycle():
    """A tree and an added cycle"""
    dig = balanced_tree()
    dig.add_edge(4, 1, value=5)

    dig.edges[(0, 1)]["value"] = dig.edges[(0, 1)]["value"] - 5
    dig.edges[(4, 10)]["value"] = dig.edges[(4, 10)]["value"] - 5
    return dig


def get_example_with_node_cycle():
    """A tree with an added cycle from one node to itself"""
    dig = balanced_tree()
    dig.add_edge(4, 4, value=2)
    return dig


example_fun_list.extend([get_example_with_simple_cycle, get_example_with_node_cycle])  # type: ignore


def get_digraph_with_edge_values_from_tree(tree: nx.DiGraph):
    """Get a directed graph with edge values based on a tree

    Args:
        tree: a directed graph
    """

    dig = tree.copy()

    def get_value_of_tree_node(node):
        if dig.out_degree(node) == 0:
            if isinstance(node, str):
                value = 1
            else:
                value = node
        elif "value" in dig.nodes[node]:
            value = dig.nodes[node]["value"]
        else:
            value = sum(
                get_value_of_tree_node(child_node)
                for child_node in dig.successors(node)
            )
        dig.nodes[node]["value"] = value
        return value

    for node in dig:
        get_value_of_tree_node(node)

    nx.get_node_attributes(dig, "value")

    for edge in dig.edges:
        dig.edges[edge]["value"] = dig.nodes[edge[1]]["value"]

    nx.get_edge_attributes(dig, "value")
    return dig


def add_cross_edge(dig: nx.DiGraph, cross_edge: tuple, edge_value: float):
    """Add an edge to a directed graph

    After adding the edge, the added value is propagated
    on the left and right
    """
    dig.add_edge(cross_edge[0], cross_edge[1])
    dig.edges[cross_edge]["value"] = edge_value
    right_node = cross_edge[1]
    while dig.out_degree(right_node) > 0:
        after_node = list(dig.successors(right_node))[0]
        dig.edges[(right_node, after_node)]["value"] += edge_value
        right_node = after_node
    left_node = cross_edge[0]
    while dig.in_degree(left_node) > 0:
        before_node = list(dig.predecessors(left_node))[0]
        dig.edges[(before_node, left_node)]["value"] += edge_value
        left_node = before_node


def get_mirrored_graph(dig: nx.DiGraph, mirror_over_leaves: bool = True):
    """Get double graph by mirroring over leaves or root

    Args:
        dig: a directed graph to mirror
        mirror_over_leaves: whether to mirror over leaves or root(s)
    """
    siamese_dig = copy.deepcopy(dig)

    if mirror_over_leaves:

        def stays_constant(node_id):
            return dig.out_degree(node_id) == 0

    else:

        def stays_constant(node_id):
            return dig.in_degree(node_id) == 0

    def mirror_node(node_id):
        if stays_constant(node_id):
            return node_id
        return f"mirror_{node_id}"

    for node in dig.nodes:
        if not stays_constant(node) > 0:
            siamese_dig.add_node(mirror_node(node))
    for edge in dig.edges:
        parent_mirror = mirror_node(edge[0])
        child_mirror = mirror_node(edge[1])
        siamese_dig.add_edge(
            child_mirror, parent_mirror, value=dig.edges[edge]["value"]
        )
    return siamese_dig


def translate_iea_energy_balance(iea_df, year: int = 2020):
    """Translate energy balance data from IEA World Energy Balances

    See https://www.iea.org/data-and-statistics/data-product/world-energy-balances-highlights

    Example use:


    iea_df_full = pd.read_excel(iea_xlsx_path, sheet_name="TimeSeries_1971-2021", skiprows=1)
    country = "France"
    year = 2020
    iea_df = iea_df_full[iea_df_full["Country"]==country]
    flow_df = translate_iea_energy_balance(iea_df, year)

    Args:
        iea_df: a dataframe from the world energy balance highlights dataset
        year: year to visualize
    """
    flow_dict = {}  # type: Dict[str, Dict[str, float]]
    for product, flow, value in zip(iea_df["Product"], iea_df["Flow"], iea_df[year]):
        translate_iea_balance_item(product, flow, value, flow_dict)

    flow_df = pd.DataFrame(flow_dict).transpose()

    for carrier in [
        "Electricity",
        "Electricity, CHP and heat plants",
        "Oil products",
        "Heat",
    ]:
        carrier_in = flow_df.loc[flow_df["target"] == carrier, "value"].sum()
        carrier_out = flow_df.loc[flow_df["source"] == carrier, "value"].sum()
        carrier_losses = carrier_in - carrier_out
        if carrier_losses < 0:
            print(f"Gains for {carrier}?! Should not be")
        elif carrier_losses > carrier_in * 0.05:
            flow_df.loc[f"{carrier} losses"] = {
                "source": carrier,
                "target": "Losses (recalculated)",
                "value": carrier_losses,
            }
    return flow_df


def translate_iea_balance_item(product, flow, value, flow_dict):
    """Translate item data from product/flow to source/target structure"""
    flow = flow.split(" (")[0]
    flow_name = f"{product}: {flow}"
    ignore = product == "Total"
    try:
        value = float(value)
    except ValueError:
        print(f"Value for {flow_name} not valid: {value}")
        ignore = True
        # print(flow_name)

    if flow in ["Production", "Imports"]:
        source = f"{product} {flow}"
        target = product
        if flow == "Imports":
            # additional flow
            pass  # flow_dict[flow_name+" import"] = {"source": "Imports", "target": f"{product} {flow}", "value": value}
    elif flow in [
        "Exports",
        "Industry",
        "Transport",
        "Residential",
        "Commercial and public services",
        "Other final consumption",
    ]:
        if flow == "Exports":
            value = -value

        source = product
        target = flow  # f"{product} {flow}"
    elif flow in [
        "Oil refineries, transformation",
        "Electricity, CHP and heat plants",
    ]:
        if value > 0:
            target = product
            source = flow
        else:
            value = -value
            source = product
            target = flow
    elif flow in ["Total final consumption", "Total energy supply"]:
        ignore = True
    else:
        print(f"Unknown flow type {flow}")
        ignore = True
    if value == 0:
        ignore = True
    if not ignore:
        flow_dict[flow_name] = {"source": source, "target": target, "value": value}



"""Optimization-based layout functions

Optimization of the vertical position of each node formulated as a linear problem.
"""

import itertools
from typing import Dict

import networkx as nx
import pulp

from . import processing


def var_name(node_name: str) -> str:
    """Get a variable name valid for pulp

    Args:
        node_name: a string or other hashable type

    Returns:
        a modified node name without special characters
    """
    node_name = str(node_name)

    replace_dict = {
        "-": "(hyphen)",
        ",": "(comma)",
        "+": "(plus)",
        "[": "(leftBracket)",
        "]": "(rightBracket)",
        " ": "(space)",
        ">": "(greaterThan)",
        "/": "(slash)",
    }
    for to_repl, replacement in replace_dict.items():
        node_name = node_name.replace(to_repl, replacement)

    return node_name


def add_bendiness_objective(
    prob: pulp.LpProblem, dig: nx.DiGraph, node_pos_dict: Dict[str, float]
):
    """Add bendiness to the objective function

    Args:
        prob: the PuLP problem
        dig: the directed graph
        node_pos_dict: a dictionary of node positions

    Returns:
        the bendiness sum variable
    """
    bendiness_vars = []
    for n_from, n_to in dig.edges:
        edge_bendiness = pulp.LpVariable(
            "bendiness_" + var_name(n_from) + "_" + var_name(n_to)
        )
        bendiness_vars.append(edge_bendiness)
        prob += (
            node_pos_dict[n_from] - node_pos_dict[n_to]
        ) <= edge_bendiness, f"y_from - y_to <= Edge bendiness {n_from} {n_to}"
        prob += (
            node_pos_dict[n_to] - node_pos_dict[n_from]
        ) <= edge_bendiness, f"y_to - y_from <= Edge bendiness {n_from} {n_to}"

    bendiness_sum = pulp.LpVariable("sumBendiness")
    prob += bendiness_sum == pulp.lpSum(bend_var for bend_var in bendiness_vars)
    return bendiness_sum


def add_distance_to_center(
    prob: pulp.LpProblem, node_pos_dict: Dict[str, float], node: str
):
    """Add a variable corresponding to the absolute distance to 0

    Args:
        prob: the PuLP problem
        node_pos_dict: a dictionary of node positions
        node: the node name

    Returns:
        the absolute node position variable
    """
    abs_node_pos = pulp.LpVariable("abs(" + var_name(node) + ")")
    prob += node_pos_dict[node] <= abs_node_pos, f"x <= |x| for {node}"
    prob += -node_pos_dict[node] <= abs_node_pos, f"-x <= |x| for {node}"
    return abs_node_pos


def optimize_absolute_vertical_position_lp(dig: nx.DiGraph):
    """Optimize absolute position of nodes, where relative position is already fixed

    The formulation minimizes the distances of nodes to center (y=0)
    and the vertical distances between nodes joined by an edge

    Args:
        dig: the directed graph

    Returns:
        None. The node positions ("y" attribute) are updated in the graph
    """
    # Create the 'prob' variable to contain the problem data
    prob = pulp.LpProblem("SugiyamaLP", pulp.LpMinimize)

    # Variables
    layers = processing.get_layers(dig)

    layer_node_dict = {
        lay: processing.get_layer_nodes(dig, lay, key_name="vertical_position")
        for lay in layers
    }

    node_pos_dict = {}
    center_distances = []
    for lay in layers:
        layer_nodes = layer_node_dict[lay]

        for i_n, node in enumerate(layer_nodes):
            node_pos_dict[node] = pulp.LpVariable(var_name(node))
            abs_node_pos = add_distance_to_center(prob, node_pos_dict, node)
            center_distances.append(abs_node_pos)

            if i_n > 0:
                prev_n = layer_nodes[i_n - 1]
                prob += (
                    node_pos_dict[prev_n] + dig.nodes[prev_n]["max_value"]
                    <= node_pos_dict[node] - dig.nodes[node]["max_value"] / 2
                ), f"Minimum distance between nodes {prev_n} {node}"

    center_dist_sum = pulp.LpVariable("sumDistancesToCenter")
    prob += center_dist_sum == pulp.lpSum(cent_dist for cent_dist in center_distances)

    bendiness_sum = add_bendiness_objective(prob, dig, node_pos_dict)

    lambda_bendiness = 2.0
    prob += (
        center_dist_sum + lambda_bendiness * bendiness_sum,
        "Minimize sum of distances to center plus bendiness",
    )

    # The problem is solved using PuLP's choice of Solver
    prob.solve(pulp.PULP_CBC_CMD(msg=1))

    for node, opt_var in node_pos_dict.items():
        dig.nodes[node]["y"] = opt_var.varValue


def optimize_vertical_position_milp(dig: nx.DiGraph, verbose: bool = False) -> None:
    """Optimize absolute and relative position of nodes

    The formulation minimizes the distances of nodes to center (y=0)
    and the vertical distances between nodes joined by an edge

    Args:
        dig: the directed graph
        verbose: whether to print the problem

    Returns:
        None. The node positions ("y" attribute) are updated in the graph
    """
    # Parameters
    lambda_bendiness = 2.0
    lambda_crossings = 200.0
    min_dist_between_nodes = 5
    # Create the 'prob' variable to contain the problem data
    prob = pulp.LpProblem("SugiyamaLP", pulp.LpMinimize)

    # Variables
    layers = processing.get_layers(dig)

    layer_node_dict = {lay: processing.get_layer_nodes(dig, lay) for lay in layers}

    node_pos_dict = {}
    node_rel_pos_dict = {}
    crossing_vars = []

    center_distances = []

    for lay in layers:
        src_layer_nodes = layer_node_dict[lay]

        for node in src_layer_nodes:
            node_pos_dict[node] = pulp.LpVariable(var_name(node))
            abs_node_pos = add_distance_to_center(prob, node_pos_dict, node)
            center_distances.append(abs_node_pos)

        for i_n1, i_n2 in itertools.permutations(range(len(src_layer_nodes)), 2):
            n_1 = src_layer_nodes[i_n1]
            n_2 = src_layer_nodes[i_n2]
            node_rel_pos_dict[(n_1, n_2)] = pulp.LpVariable(
                var_name(n_1) + "_above_" + var_name(n_2), cat="Binary"
            )

            # if n1 is above n2, yn1 - yn2 should be above 0 and larger than some distance
            prob += (
                node_pos_dict[n_1]
                - node_pos_dict[n_2]
                - dig.nodes[n_1]["max_value"] / 2
                - dig.nodes[n_2]["max_value"] / 2
                - min_dist_between_nodes
                + 1000 * (1 - node_rel_pos_dict[(n_1, n_2)])
                >= 0
            )

        # constraint: n1 above n2 or n2 above n1 for any pair of nodes n1, n2
        for i_n1, i_n2 in itertools.combinations(range(len(src_layer_nodes)), 2):
            n_1 = src_layer_nodes[i_n1]
            n_2 = src_layer_nodes[i_n2]
            prob += node_rel_pos_dict[(n_1, n_2)] + node_rel_pos_dict[(n_2, n_1)] == 1

    # crossings
    for src_lay, next_lay in zip(layers[:-1], layers[1:]):
        src_layer_nodes = processing.get_layer_nodes(dig, src_lay)
        next_layer_nodes = processing.get_layer_nodes(dig, next_lay)
        layer_edges = [
            edge
            for edge in dig.edges
            if edge[0] in src_layer_nodes and edge[1] in next_layer_nodes
        ]
        for i_e1, i_e2 in itertools.combinations(range(len(layer_edges)), 2):
            n1a, n1b = layer_edges[i_e1]
            n2a, n2b = layer_edges[i_e2]

            if not (n1a == n2a) and not (n1b == n2b):
                cross_var_name = (
                    var_name(n1a)
                    + "_"
                    + var_name(n1b)
                    + "_crosses_"
                    + var_name(n2a)
                    + "_"
                    + var_name(n2b)
                )
                cross_var = pulp.LpVariable(
                    cross_var_name,
                    cat="Binary",
                )  # 1 if the two edges cross
                prob += (
                    cross_var
                    >= node_rel_pos_dict[(n1a, n2a)] + node_rel_pos_dict[(n2b, n1b)] - 1
                )
                prob += (
                    cross_var
                    >= node_rel_pos_dict[(n2a, n1a)] + node_rel_pos_dict[(n1b, n2b)] - 1
                )
                crossing_vars.append(cross_var)

    center_dist_sum = pulp.LpVariable("sumDistancesToCenter")
    prob += center_dist_sum == pulp.lpSum(cent_dist for cent_dist in center_distances)

    n_crossings = pulp.LpVariable("numberOfEdgeCrossings")
    prob += n_crossings == pulp.lpSum(cr for cr in crossing_vars)

    bendiness_sum = add_bendiness_objective(prob, dig, node_pos_dict)

    prob += (
        center_dist_sum
        + lambda_bendiness * bendiness_sum
        + lambda_crossings * n_crossings,
        "Minimize sum of crossings plus distances to center plus bendiness",
    )

    # The problem is solved using PuLP's choice of Solver
    # prob.solve()
    prob.solve(pulp.PULP_CBC_CMD(msg=1))

    if verbose:
        # Each of the variables is printed with it's resolved optimum value
        for opt_var in prob.variables():
            print(opt_var.name, "=", opt_var.varValue)

    for node, opt_var in node_pos_dict.items():
        dig.nodes[node]["y"] = opt_var.varValue
        dig.nodes[node]["vertical_position"] = 0

    for lay in layers:
        src_layer_nodes = layer_node_dict[lay]
        for i_n1, i_n2 in itertools.permutations(range(len(src_layer_nodes)), 2):
            n_1 = src_layer_nodes[i_n1]
            n_2 = src_layer_nodes[i_n2]
            dig.nodes[n_1]["vertical_position"] += node_rel_pos_dict[
                (n_1, n_2)
            ].varValue
            if verbose:
                print((n_1, n_2, node_rel_pos_dict[(n_1, n_2)].varValue))

    return prob



"""Drawing Sankey diagrams with Bokeh

Bokeh allows interactions to be included in plots: zooming, panning, showing things on hover etc.
"""

from typing import Any, Dict, Optional

import networkx as nx
import pandas as pd
from bokeh.models import ColumnDataSource, HoverTool
from bokeh.plotting import figure

from . import draw


def get_bokeh_kwargs(pyplot_kwargs: Dict[str, Any]) -> dict:
    """Translate pyplot keyword arguments in bokeh keyword arguments

    Args:
        pyplot_kwargs (Dict[str, Any]): A dictionary of pyplot keyword arguments.

    Returns:
        bk_kwargs: A dictionary of bokeh keyword arguments.
    """
    bk_kwargs = {}
    kw_translate_dict = {"lw": "line_width"}
    for plt_kw, kw_val in pyplot_kwargs.items():
        bk_kw = kw_translate_dict.get(plt_kw, plt_kw)
        bk_kwargs[bk_kw] = kw_val
    return bk_kwargs


def draw_sankey_bokeh(dig: nx.DiGraph, draw_config: Optional[draw.DrawConfig] = None):
    """The actual drawing of a Sankey plot in bokeh

    Args:
        dig
        draw_config

    Returns:
        figure
    """

    if draw_config is None:
        draw_config = draw.DrawConfig()

    diagram = draw.get_sankey_diagram(dig, draw_config)

    x_range, y_range = diagram.coordinate_range

    plot = figure(
        width=800,
        height=300,
        title="",
        tools="",
        toolbar_location=None,
        match_aspect=True,
        x_range=x_range,
        y_range=y_range,
    )

    for pline in diagram.lines:
        plot.line(
            pline.x_list, pline.y_list, **get_bokeh_kwargs(draw_config.edge_edge_kw)
        )

    filled_df = pd.DataFrame(
        {"x": pline.x_list, "y": pline.y_list, "name": pline.name}
        for pline in diagram.filled
    )
    filled_source = ColumnDataSource(filled_df)
    fill_renderer = plot.patches(
        source=filled_source,
        xs="x",
        ys="y",
        **get_bokeh_kwargs(draw_config.edge_fill_kw)
    )
    fill_hover = HoverTool(renderers=[fill_renderer], tooltips=[("", "@name")])
    plot.add_tools(fill_hover)

    node_label_df = pd.DataFrame(
        [
            {"x": label.x_txt, "y": label.y_txt, "text": label.text}
            for label in diagram.labels
            if label.category == "node_label"
        ]
    )
    node_label_source = ColumnDataSource(node_label_df)
    plot.text(
        source=node_label_source,
        x="x",
        y="y",
        text="text",
        **draw_config.node_label_kwargs
    )

    edge_label_df = pd.DataFrame(
        [
            {"x": label.x_txt, "y": label.y_txt, "text": label.text}
            for label in diagram.labels
            if label.category == "edge_label"
        ]
    )
    edge_label_source = ColumnDataSource(edge_label_df)
    plot.text(
        source=edge_label_source,
        x="x",
        y="y",
        text="text",
        **draw_config.edge_label_kwargs
    )

    plot.xgrid.grid_line_color = None
    plot.ygrid.grid_line_color = None
    plot.axis.visible = False
    return plot



"""Draw (Sankey diagram or node link diagrams)"""

from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Tuple, Union

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np


def get_default_text_kwargs(label_category: str, library: str) -> Dict[str, Any]:
    """Get default keyword arguments.

    Args:
        label_category: The category of the label.
            One of "node_label" or "edge_label".
        library: The library being used
            "matplotlib" or "bokeh"

    Returns:
        A dictionary with the default keyword argument, e.g. {"fontsize": 10}.
    """
    if label_category == "node_label":
        font_size: Union[int, str] = 10
    elif label_category == "edge_label":
        font_size = 8
    else:
        raise NotImplementedError("Unknown label category")

    if library.lower() == "matplotlib":
        font_size_attr = "fontsize"
        ha_attr = "ha"
    elif library.lower() == "bokeh":
        font_size_attr = "text_font_size"
        font_size = f"{font_size}px"
        ha_attr = "text_align"
    else:
        raise NotImplementedError(
            "Unknown library, currently implemented are Matplotlib and Bokeh."
        )

    txt_kwargs = {font_size_attr: font_size, ha_attr: "center"}
    return txt_kwargs


@dataclass
class DrawConfig:
    """Options for drawing

    Args:
        n_segments
        dx_node: node half width
        dx_arrow
    """

    n_segments: int = 100
    link_geometry: str = "cubic_spline"
    write_node_names: bool = True
    node_name_dy_frac: float = (
        0.1  # vertical distance between node center and node label, as a fraction of dy_node
    )
    library: str = "Matplotlib"
    node_label_kwargs: Optional[Dict[str, Any]] = None
    write_edge_values: bool = True
    edge_label_kwargs: Optional[Dict[str, Any]] = None
    edge_label_format: str = "{}"
    dx_node: float = 0.2
    dx_arrow: float = 0.1
    edge_edge_kw: Optional[Dict[str, Any]] = None
    node_edge_kw: Optional[Dict[str, Any]] = None
    edge_fill_kw: Optional[Dict[str, Any]] = None
    node_fill_kw: Optional[Dict[str, Any]] = None
    debug_visuals: bool = False
    link_color_attribute: Optional[str] = None

    def __post_init__(self):
        """Default kwargs in post-init"""
        if self.edge_edge_kw is None:
            self.edge_edge_kw = {"color": "gray"}
        if self.node_edge_kw is None:
            self.node_edge_kw = {"color": "gray"}
        if self.edge_fill_kw is None:
            self.edge_fill_kw = {"color": "deeppink", "alpha": 0.2, "lw": 0}
        if self.node_fill_kw is None:
            self.node_fill_kw = {"color": "deeppink", "alpha": 0.2, "lw": 0}
        if self.node_label_kwargs is None:
            self.node_label_kwargs = get_default_text_kwargs("node_label", self.library)
        if self.edge_label_kwargs is None:
            self.edge_label_kwargs = get_default_text_kwargs("edge_label", self.library)


def plot_node_positions(dig: nx.DiGraph, pos_var="y"):
    """Plot node positions

    Args:
        dig: a directed graph with node attributes "layer" and pos_var (by default "y")
        pos_var: a string determining which node attribute to use as y-coordinate
    """

    plt.figure()
    color_cycle = plt.rcParams["axes.prop_cycle"].by_key()["color"]

    node_colors = [
        color_cycle[dig.nodes[node]["layer"] % len(color_cycle)] for node in dig.nodes
    ]
    node_positions = {
        node: (dig.nodes[node]["layer"], dig.nodes[node][pos_var]) for node in dig.nodes
    }
    nx.draw(dig, node_color=node_colors, pos=node_positions, with_labels=True)

    for node in dig.nodes:
        x_node = dig.nodes[node]["layer"]
        y_node = dig.nodes[node][pos_var]
        dy_node = dig.nodes[node]["max_value"]
        plt.plot(
            [x_node, x_node], [y_node - dy_node / 2, y_node + dy_node / 2], color="gray"
        )


def cubic_spline_link(
    x_from_to: Optional[np.ndarray] = None,
    y_from_to: Optional[np.ndarray] = None,
    n_segments: int = 100,
):
    """Draw a (horizontally oriented) cubic spline with 0 slope from a point to another point

    Args:
        x_from_to: a numpy array with two x coordinates
        y_from_to: a numpy array with two y coordinates
        n_segments: number of straight segments with which to draw the spline

    Returns:
        two numpy arrays
    """
    if x_from_to is None:
        x_from_to = np.array([0, 1])
    if y_from_to is None:
        y_from_to = np.array([0, 1])

    assert n_segments >= 1, "There should be at least one segment"
    x_t = np.linspace(0, 1, n_segments + 1)
    y_coord = y_from_to[0] + (y_from_to[1] - y_from_to[0]) * (x_t**2 * (3 - 2 * x_t))
    x_coord = x_from_to[0] + (x_from_to[1] - x_from_to[0]) * x_t

    return x_coord, y_coord


def get_node_position(dig: nx.DiGraph, node: str) -> tuple:
    """Get node position

    Args:
        dig
        node

    Returns:
        node_x: node center x coordinate
        node_y: node (bottom) y coordinate
        d_y: node height
    """
    node_x = dig.nodes[node]["layer"]
    d_y = dig.nodes[node]["max_value"]
    node_y = dig.nodes[node]["y"] - d_y / 2
    return node_x, node_y, d_y


@dataclass
class Polyline:
    """A polyline or broken line"""

    x_list: List[float]
    y_list: List[float]
    name: str = ""
    color_attribute_value: Optional[str] = None

    @property
    def n_points(self):
        """Number of points"""
        n_x = len(self.x_list)
        n_y = len(self.y_list)
        if n_x == n_y:
            return n_x
        else:
            raise ValueError(
                "x and y-coordinates do not have the same number of points"
            )

    @property
    def is_empty(self):
        """Check whether polyline has no point"""
        return self.n_points == 0

    @property
    def coordinate_range(self):
        """Min and max or x- and y-coordinates"""
        x_range = min(self.x_list), max(self.x_list)
        y_range = min(self.y_list), max(self.y_list)
        return x_range, y_range

    def start_touches_end(self, other):
        """Does this polyline start where the other polyline ends"""
        return (self.x_list[0] == other.x_list[-1]) and (
            self.y_list[0] == other.y_list[-1]
        )

    def append_to(self, other):
        """Append this polyline to another one"""
        return Polyline(
            other.x_list + self.x_list,
            other.y_list + self.y_list,
            other.name + " + " + self.name,
        )

    def get_center(self):
        """Return coordinates of polyline barycenter"""
        x_center = np.mean(np.array(self.x_list))
        y_center = np.mean(np.array(self.y_list))
        return x_center, y_center


def concatenate_polylines(polylines: List[Polyline]) -> List[Polyline]:
    """Concatenate polylines, if their ends touch

    Args:
        polylines: to concatenate

    Returns:
        new_plines: list of polylines, fewer than the original one
    """
    new_plines = []
    candidate_pline = None
    polylines = [pline for pline in polylines if not pline.is_empty]
    for pline in polylines:
        if candidate_pline is None:
            # first polyline
            candidate_pline = pline
        elif not pline.start_touches_end(candidate_pline):
            # disjoint polylines: add candidate_pline to list, make new one candidate
            new_plines.append(candidate_pline)
            candidate_pline = pline
        else:
            candidate_pline = pline.append_to(candidate_pline)
    assert candidate_pline is not None
    new_plines.append(candidate_pline)
    return new_plines


def get_polygon(polylines: List[Polyline]) -> Polyline:
    """Concatenate polylines even if their ends do not touch (for filling)"""
    polygon = polylines[0]
    for pline in polylines[1:]:
        polygon = pline.append_to(polygon)
    return polygon


@dataclass
class Label:
    """A text element in the diagram"""

    x_txt: float
    y_txt: float
    text: str
    category: str = ""


@dataclass
class Diagram:
    """A container for the elements of a diagram, independently of the plotting library

    Attributes:
        lines: polylines, not filled
        filled: polylines, filled
        labels: text labels
    """

    lines: List[Polyline] = field(default_factory=lambda: [])
    filled: List[Polyline] = field(default_factory=lambda: [])
    labels: List[Label] = field(default_factory=lambda: [])
    color_dict: Dict[str, str] = field(default_factory=lambda: {})

    @property
    def coordinate_range(self):
        """Min and max or x- and y-coordinates"""
        (x_min, x_max), (y_min, y_max) = self.lines[0].coordinate_range
        for pline in self.lines[1:]:
            (pl_x_min, pl_x_max), (pl_y_min, pl_y_max) = pline.coordinate_range
            x_min = min(x_min, pl_x_min)
            x_max = max(x_max, pl_x_max)
            y_min = min(y_min, pl_y_min)
            y_max = max(y_max, pl_y_max)

        x_range = (x_min, x_max)
        y_range = (y_min, y_max)
        return x_range, y_range


def out_arrow_coordinates(dig, node, draw_config):
    """Get coordinates of out-flow arrow"""
    node_x, node_y, d_y = get_node_position(dig, node)
    dx_node, dx_arrow = draw_config.dx_node, draw_config.dx_arrow
    draw_out_arrow = dig.out_degree(node) == 0 and dx_arrow > 0

    if draw_out_arrow:
        x_arrow_out = [
            node_x + dx_node,
            node_x + dx_node + dx_arrow,
            node_x + dx_node,
        ]
        y_arrow_out = [node_y, node_y + d_y / 2, node_y + d_y]
    else:
        x_arrow_out = []
        y_arrow_out = []
    return Polyline(x_arrow_out, y_arrow_out, name="outflow arrow")


def in_arrow_coordinates(dig, node, draw_config):
    """Get coordinates of in-flow arrow"""
    node_x, node_y, d_y = get_node_position(dig, node)
    dx_node, dx_arrow = draw_config.dx_node, draw_config.dx_arrow
    draw_in_arrow = dig.in_degree(node) == 0 and dx_arrow > 0
    if draw_in_arrow:
        x_arrow_in = [
            node_x - dx_node,
            node_x - dx_node + dx_arrow,
            node_x - dx_node,
        ]
        y_arrow_in = [node_y + d_y, node_y + d_y / 2, node_y]

    else:
        x_arrow_in = []
        y_arrow_in = []
    return Polyline(x_arrow_in, y_arrow_in, name="inflow arrow")


@dataclass
class NodeRepresentation:
    """Geometric representation of a node"""

    polylines: List[Polyline]  # edges
    polygon: Optional[Polyline]  # filled
    label: Optional[Label]
    color_attribute_value: Optional[str] = None

    def __post_init__(self):
        col_attr_val = self.color_attribute_value
        if self.polygon is not None and col_attr_val is not None:
            for poly in self.polylines:
                poly.color_attribute_value = col_attr_val
            self.polygon.color_attribute_value = col_attr_val


def node_dx(draw_config: nx.DiGraph, node: str):
    """Get node width"""
    dx_node = draw_config.dx_node
    if "dummy" in str(node).lower():
        dx_node = 0
    return dx_node


def get_node_geometry(
    dig: nx.DiGraph, node: str, draw_config: DrawConfig, color_attribute_value=None
) -> NodeRepresentation:
    """Get geometry of a Sankey diagram for one node

    Args:
        dig
        node: node identifier, typically a str
    """
    node_x, node_y, _ = get_node_position(dig, node)

    dx_node = node_dx(draw_config, node)

    node_label = None
    if dx_node > 0 and dig.nodes[node]["max_value"] > 0:
        # plot "node", if any width
        bottom_line = Polyline(
            [node_x - dx_node, node_x + dx_node], [node_y, node_y], "bottom line"
        )
        y_top = node_y + dig.nodes[node]["max_value"]
        top_line = Polyline(
            [node_x + dx_node, node_x - dx_node], [y_top, y_top], "top line"
        )

        out_arrow = out_arrow_coordinates(dig, node, draw_config=draw_config)
        in_arrow = in_arrow_coordinates(dig, node, draw_config=draw_config)

        node_plines = [bottom_line, out_arrow, top_line, in_arrow]
        plot_plines = concatenate_polylines(node_plines)

        fill_poly = get_polygon(plot_plines)
        fill_poly.name = str(node)

        if draw_config.write_node_names and "dummy" not in str(node).lower():
            y_label = (
                dig.nodes[node]["y"]
                + dig.nodes[node]["max_value"] * draw_config.node_name_dy_frac
            )
            node_label = Label(node_x, y_label, str(node), "node_label")

    else:
        plot_plines, fill_poly = [], None
    return NodeRepresentation(
        plot_plines, fill_poly, node_label, color_attribute_value=color_attribute_value
    )


@dataclass
class LinkRepresentation:
    """Geometric representation of a link"""

    polylines: List[Polyline]  # edges
    polygon: Polyline  # filled
    label: Optional[Label]
    value: float
    color_attribute_value: Optional[str] = None

    def __post_init__(self):
        col_attr_val = self.color_attribute_value
        if col_attr_val is not None:
            for poly in self.polylines:
                poly.color_attribute_value = col_attr_val
            self.polygon.color_attribute_value = col_attr_val


def get_edge_color_value(dig, edge, draw_config):
    """Get color value of an edge"""
    if draw_config.link_color_attribute is not None:
        edge_color_val = dig.edges[edge][draw_config.link_color_attribute]
    else:
        edge_color_val = None
    return edge_color_val


def backward_link_curve(
    x_from_to,
    y_from_to,
    edge_val: float,
    below: bool,
    draw_config: Optional[DrawConfig] = None,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Get geometry of backward link (as arises in the case of cycles)

    One half-ellipse to turn backwards, a spline and another half-ellipse to get forward into the node again

    Args:
        x_from_to: x coordinates of the two nodes
        y_from_to: y coordinates of the two nodes
        edge_val: Sankey diagram edge value
            the ellipse height will be proportional to it
        below: true for the lower edge of the link, false for the upper edge
    """
    if draw_config is None:
        draw_config = DrawConfig()
    if below:
        x_radius = 0.4
        y_radius = 2.0 * edge_val
    else:
        x_radius = 0.2
        y_radius = 1.0 * edge_val
    angle = np.linspace(-np.pi / 2, np.pi / 2, draw_config.n_segments)
    x_first_half_circle = x_from_to[0] + x_radius * np.cos(angle)
    y_first_half_circle = y_from_to[0] + y_radius * (1 + np.sin(angle))
    x_second_half_circle = x_from_to[1] - x_radius * np.cos(angle)
    y_second_half_circle = y_from_to[1] + y_radius * (1 + np.sin(angle[::-1]))
    x_link, y_link = cubic_spline_link(
        np.array([x_first_half_circle[-1], x_second_half_circle[0]]),
        np.array([y_first_half_circle[-1], y_second_half_circle[0]]),
        n_segments=draw_config.n_segments,
    )
    x_curve = np.concatenate([x_first_half_circle, x_link, x_second_half_circle])

    y_curve = np.concatenate([y_first_half_circle, y_link, y_second_half_circle])
    if not below:
        y_curve = y_curve + edge_val
    return x_curve, y_curve


def get_link_geometry(
    dig: nx.DiGraph, node, next_node, y_from, draw_config: DrawConfig
) -> LinkRepresentation:
    """Geometry of link between two nodes as part of a Sankey diagram"""

    dx_node = node_dx(draw_config, node)
    dx_node_next = node_dx(draw_config, next_node)
    x_mid, _, _ = get_node_position(dig, node)

    edge_val = dig.edges[node, next_node]["value"]
    x_next = dig.nodes[next_node]["layer"]
    y_next = dig.nodes[next_node]["y_in"]

    x_from_to = np.array([x_mid + dx_node, x_next - dx_node_next])
    y_from_to = np.array([y_from, y_next])
    if x_from_to[1] < x_from_to[0]:
        # special case of backward link
        x_below, y_below = backward_link_curve(
            x_from_to, y_from_to, edge_val, below=True, draw_config=draw_config
        )
        x_above, y_above = backward_link_curve(
            x_from_to, y_from_to, edge_val, below=False, draw_config=draw_config
        )
        # y_above = y_below + edge_val
    elif draw_config.link_geometry == "line":
        x_below = x_from_to
        y_below = y_from_to
        y_above = y_from_to + edge_val
        x_above = x_below
    elif draw_config.link_geometry == "cubic_spline":
        x_below, y_below = cubic_spline_link(
            x_from_to, y_from_to, n_segments=draw_config.n_segments
        )
        y_above = y_below + edge_val
        x_above = x_below

    x_fill = np.concatenate([x_below, x_above[::-1]])
    y_fill = np.concatenate([y_below, y_above[::-1]])

    edge_plines = [
        Polyline(x_below.tolist(), y_below.tolist()),
        Polyline(x_above.tolist(), y_above.tolist()),
    ]
    edge_name = f"{node} -> {next_node}"
    edge_poly = Polyline(x_fill.tolist(), y_fill.tolist(), name=edge_name)

    dig.nodes[next_node]["y_in"] += edge_val

    if draw_config.write_edge_values:
        label_str = draw_config.edge_label_format.format(edge_val)
        edge_label = Label(np.mean(x_fill), np.mean(y_fill), label_str, "edge_label")
    else:
        edge_label = None

    edge_color_val = get_edge_color_value(dig, (node, next_node), draw_config)

    return LinkRepresentation(
        edge_plines,
        edge_poly,
        edge_label,
        edge_val,
        color_attribute_value=edge_color_val,
    )


def get_sankey_diagram(dig: nx.DiGraph, draw_config=None) -> Diagram:
    """Determine the diagram geometry"""

    def node_y_position(nod):
        """Node vertical position"""
        return dig.nodes[nod]["vertical_position"]

    def node_layer_and_y_position(nod):
        """Node vertical position"""
        return (-dig.nodes[nod]["layer"], dig.nodes[nod]["vertical_position"])

    if draw_config is None:
        draw_config = DrawConfig()

    for node in dig.nodes:
        dig.nodes[node]["y_in"] = dig.nodes[node]["y"] - dig.nodes[node]["in_value"] / 2

    sorted_nodes = sorted(
        dig.nodes,
        key=node_layer_and_y_position,
    )
    diagram = Diagram()
    color_attribute_values = []
    for node in sorted_nodes:
        node_color_attribute_values = []

        x, node_y, dy = get_node_position(dig, node)

        sorted_successors = sorted(dig.successors(node), key=node_y_position)
        for n_next in sorted_successors:
            link_repr = get_link_geometry(
                dig,
                node,
                n_next,
                node_y,
                draw_config,
            )
            node_y += link_repr.value
            diagram.lines.extend(link_repr.polylines)
            diagram.filled.append(link_repr.polygon)
            if link_repr.label is not None:
                diagram.labels.append(link_repr.label)

            if link_repr.color_attribute_value is not None:
                color_attribute_values.append(link_repr.color_attribute_value)
                node_color_attribute_values.append(link_repr.color_attribute_value)

        predecessors = dig.predecessors(node)
        for node_pred in predecessors:
            edge_color_val = get_edge_color_value(dig, (node_pred, node), draw_config)
            if edge_color_val is not None:
                node_color_attribute_values.append(edge_color_val)

        node_color_attribute_values = list(set(node_color_attribute_values))
        # we give node a color if all incoming and outcoming edges have the same color
        if len(node_color_attribute_values) == 1:
            node_color_value = node_color_attribute_values[0]
        elif dig.out_degree(node) > 0 and dig.in_degree(node) > 0:
            # non-terminal node, allow to have color
            node_color_value = "node"
        else:
            # terminal node, melt with edge
            node_color_value = None

        node_repr = get_node_geometry(
            dig, node, draw_config, color_attribute_value=node_color_value
        )

        diagram.lines.extend(node_repr.polylines)
        if node_repr.polygon is not None:
            diagram.filled.append(node_repr.polygon)
        if node_repr.label is not None:
            diagram.labels.append(node_repr.label)

    if draw_config.link_color_attribute is not None:
        color_attribute_values = sorted(list(set(color_attribute_values)))
        color_cycler = plt.rcParams["axes.prop_cycle"].by_key()["color"]
        diagram.color_dict = {
            attr_val: color_cycler[i_val]
            for i_val, attr_val in enumerate(color_attribute_values)
        }

    return diagram


def draw_sankey(dig, ax=None, draw_config=None):
    """The actual drawing in Matplotlib

    Args:
        dig: ready-to-plot networkx.DiGraph
        ax: a matplotlib axis

    """

    if draw_config is None:
        draw_config = DrawConfig()

    diagram = get_sankey_diagram(dig, draw_config)

    if ax is None:
        _, ax = plt.subplots()

    for pline in diagram.lines:
        edge_edge_kw = dict(draw_config.edge_edge_kw)
        if pline.color_attribute_value is not None:
            if pline.color_attribute_value == "node":
                edge_edge_kw = dict(draw_config.node_edge_kw)
            else:
                edge_edge_kw["color"] = diagram.color_dict[pline.color_attribute_value]
        ax.plot(pline.x_list, pline.y_list, **edge_edge_kw)

    for pline in diagram.filled:
        fill_kw = dict(draw_config.edge_fill_kw)
        if pline.color_attribute_value is not None:
            if pline.color_attribute_value == "node":
                fill_kw = dict(draw_config.node_fill_kw)
            else:
                fill_kw["color"] = diagram.color_dict[pline.color_attribute_value]

        ax.fill(pline.x_list, pline.y_list, **fill_kw)

    for label in diagram.labels:
        if label.category == "node_label":
            label_kwargs = draw_config.node_label_kwargs
        elif label.category == "edge_label":
            label_kwargs = draw_config.edge_label_kwargs
        else:
            raise ValueError("Unknown label category {label.category}")
        ax.text(label.x_txt, label.y_txt, label.text, **label_kwargs)

    # legend
    if draw_config.link_color_attribute is not None:
        for col_val, color in diagram.color_dict.items():
            edge_fill_kw = dict(draw_config.edge_fill_kw)
            edge_fill_kw["color"] = color
            ax.fill([], [], label=col_val, **edge_fill_kw)

        ax.legend(bbox_to_anchor=(0.5, -0.1), loc="upper center", ncol=2)

    ax.axis(False)
    return ax


def draw_edges(
    dig: nx.DiGraph,
    pos: Dict[Any, Tuple[float, float]],
    ax=None,
    plot_kwargs=None,
    draw_config: Optional[DrawConfig] = None,
    x_node_width: float = 0.0,
):
    """Draw edges of a directed graph, by default using cubic splines"""
    if ax is None:
        _, ax = plt.subplots()
    if plot_kwargs is None:
        plot_kwargs = {"color": "k"}
    if draw_config is None:
        draw_config = DrawConfig()
    for edge in dig.edges:
        p0 = pos[edge[0]]
        p1 = pos[edge[1]]
        x_from_to = np.array([p0[0] + x_node_width, p1[0] - x_node_width])
        y_from_to = np.array([p0[1], p1[1]])

        if edge[0] == edge[1]:
            edge_xs, edge_ys = backward_link_curve(
                x_from_to,
                y_from_to,
                edge_val=0.5,
                draw_config=draw_config,
                below=True,
            )
        elif draw_config.link_geometry == "line":
            edge_xs, edge_ys = x_from_to, y_from_to
        else:
            edge_xs, edge_ys = cubic_spline_link(
                x_from_to, y_from_to, n_segments=draw_config.n_segments
            )
        ax.plot(edge_xs, edge_ys, **plot_kwargs)



"""Processing directed graphs"""

import itertools
from dataclasses import dataclass
from typing import Optional

import networkx as nx
import numpy as np
import pandas as pd

from . import draw, optim


@dataclass
class LayoutConfig:
    """Configuration for graph layout creation

    Attributes:
        verbose
        align: for layer assignment, one of "left" and "right"
        justify: for layer assignment
        vertical_positioning: one of "barycenter_heuristic", "lp", "milp"
    """

    verbose: bool = False
    align: str = "right"
    justify: bool = True
    vertical_positioning: str = "barycenter_heuristic"
    barycenter_heuristic_n_sweep_min: int = 2
    barycenter_heuristic_n_sweep_max: int = 6


def get_graph_from_df(flow_df: pd.DataFrame) -> nx.DiGraph:
    """Get directed graph from dataframe containing flow data

    Args:
        flow_df: a dataframe with columns "source", "target" and "value", one row per flow

    Returns:
        a networkx digraph
    """
    dig = nx.DiGraph()
    node_names = list(set(flow_df["source"]).union(set(flow_df["target"])))
    dig.add_nodes_from(node_names)
    dig.add_edges_from(
        zip(
            flow_df["source"],
            flow_df["target"],
            [{"value": val} for val in flow_df["value"]],
        )
    )
    other_columns = set(flow_df.columns) - set(["source", "target", "value"])

    for col_name in other_columns:
        for src, target, col_val in zip(
            flow_df["source"], flow_df["target"], flow_df[col_name]
        ):
            dig.edges[(src, target)][col_name] = col_val

    calculate_node_values(dig)

    return dig


def get_df_from_graph(dig: nx.DiGraph) -> pd.DataFrame:
    """Return a dataframe from a directed graph

    Args:
        dig: a directed graph with edge attribute "value"

    Returns:
        a dataframe with columns "source", "target" and "value", one row per flow
    """
    edge_dict = {
        str(edge): {
            "source": edge[0],
            "target": edge[1],
            "value": dig.edges[edge]["value"],
        }
        for edge in dig.edges
    }
    flow_df = pd.DataFrame(edge_dict).transpose()
    return flow_df


def check_digraph_before_processing(dig: nx.DiGraph):
    """Check that a directed graph is fit for processing and drawing

    Args:
        dig: a directed graph with edge attribute "value"

    Raises:
        ValueError: if the directed graph is not a directed acyclic graph,
            or if edges do not have a "value" attribute
    """
    if not isinstance(dig, nx.DiGraph):
        raise ValueError(f"Should be a directed graph, but is {type(dig)}")
    if not nx.is_directed_acyclic_graph(dig):
        dig.cycle_edges = {}
        simple_cycles = list(nx.simple_cycles(dig))
        print("Simple cycles:", simple_cycles)
        while len(simple_cycles) > 0:
            cycle = simple_cycles[0]
            if len(cycle) == 1:
                cycle_edges = [(cycle[0], cycle[0])]
            else:
                cycle_edges = [
                    (cycle[i_edge], cycle[(i_edge + 1) % len(cycle)])
                    for i_edge in range(len(cycle))
                ]
            # removing the edge with the lowest value
            edge = min(cycle_edges, key=lambda edge: dig.edges[edge]["value"])
            print(f"Temporarily opening cycle {cycle} by removing edge {edge}")
            dig.cycle_edges[edge] = dig.edges[edge]
            dig.remove_edge(*edge)
            simple_cycles = list(nx.simple_cycles(dig))

        # raise ValueError(
        #    f"Should be a directed acyclic graph, but there seem to be cycles: {simple_cycles}"
        # )
    for edge in dig.edges:
        if "value" not in dig.edges[edge]:
            raise ValueError(f"No 'value' attribute for {edge}")
    print("Check directed graph before processing: seems OK")


def calculate_node_values(dig: nx.DiGraph, max_imbalance=0.05):
    """Calculate "node values" by summing edge values

    Args:
        dig: a directed graph where edges have a "value" attribute, modified in place
        max_imbalance: value above which a node is considered imbalanced and a warning is printed

    Returns:
        None. The directed graph is modified in place.
            Nodes get attributes "in_value", "out_value" and "max_value".
    """
    for node in dig.nodes:
        dig.nodes[node]["in_value"] = 0
        dig.nodes[node]["out_value"] = 0

    for n_src, n_targ, edge_data in dig.edges.data():
        edge_val = edge_data["value"]
        dig.nodes[n_targ]["in_value"] += edge_val
        dig.nodes[n_src]["out_value"] += edge_val

    for node in dig.nodes:
        in_value = dig.nodes[node]["in_value"]
        out_value = dig.nodes[node]["out_value"]
        max_value = max(in_value, out_value)
        dig.nodes[node]["max_value"] = max_value
        imbalance = abs(in_value - out_value) / (max_value + 1e-6)
        if in_value * out_value > 0 and imbalance > max_imbalance:
            print(
                f"Imbalanced node {node}: inflow {in_value:.2f} / outflow {out_value:.2f}"
            )


def assign_layers(dig: nx.DiGraph, method: Optional[LayoutConfig] = None):
    """Assign nodes to layers by length of longest directed path from each node

    This is done recursively.
    Align right: Leaves are in layer 0, parents of leaves are in layer -1 etc.
    Align left: Nodes with no parent are in layer 0, their children are in layer 1 etc.
    Align right justify:

    Args:
        dig: a networkx digraph

    Returns:
        None: the digraph is enriched with the "layer" attribute in place
    """

    if method is None:
        method = LayoutConfig()

    if method.align == "right":

        def get_nodes_to_remove(dig_copy):
            return [node for node in dig_copy if dig_copy.out_degree(node) == 0]

        layer_increment = -1
    elif method.align == "left":

        def get_nodes_to_remove(dig_copy):
            return [node for node in dig_copy if dig_copy.in_degree(node) == 0]

        layer_increment = 1
    dig_copy = dig.copy()

    layer = 0
    while dig_copy.number_of_nodes() > 0:
        nodes_to_remove = get_nodes_to_remove(dig_copy)

        if method.verbose:
            print(f"{nodes_to_remove} added to layer {layer} and removed")
        for node in nodes_to_remove:
            dig.nodes[node]["layer"] = layer
        for node in nodes_to_remove:
            dig_copy.remove_node(node)
        layer += layer_increment

    if method.justify and method.align == "right":
        for node in dig.nodes:
            if dig.in_degree(node) == 0:
                dig.nodes[node]["layer"] = layer

    if method.justify and method.align == "left":
        for node in dig.nodes:
            if dig.out_degree(node) == 0:
                dig.nodes[node]["layer"] = layer


def get_layers(dig) -> list:
    """Get layers of directed graph

    Args:
        dig: a directed graph

    Returns:
        A list of strings
    """
    layer_set = set(dig.nodes[node]["layer"] for node in dig.nodes)
    return sorted(list(layer_set))


def get_layer_nodes(dig: nx.DiGraph, lay: str, key_name: Optional[str] = None):
    """Get nodes of layer

    Args:
        dig: a directed graph
        lay: a string, name of layer
        key_name: an optional string to sort the nodes

    Returns:
        a list of strings, names of nodes
    """
    if key_name is None:

        def key_fun(node):
            return str(node)

    else:

        def key_fun(nod):
            return dig.nodes[nod][key_name]

    layer_nodes = sorted(
        [node for node in dig.nodes if dig.nodes[node]["layer"] == lay],
        key=key_fun,
    )
    return layer_nodes


def add_dummy_nodes(dig, verbose=False):
    """Add dummy node(s) for edges bridging more 2 or more layers

    Args:
        dig: a directed graph

    Returns:
        None
    """
    edges_to_break = []
    for edge in dig.edges:
        v_par, v_chi = edge
        lay_diff = dig.nodes[v_chi]["layer"] - dig.nodes[v_par]["layer"]
        if lay_diff > 1:
            edges_to_break.append(edge)

    if verbose and len(edges_to_break) > 0:
        print(f"We need a dummy node for {edges_to_break}")

    for edge in edges_to_break:
        v_par, v_chi = edge
        edge_val = dig.edges[edge]["value"]
        edge_dict = dig.edges[edge]
        lay_diff = dig.nodes[v_chi]["layer"] - dig.nodes[v_par]["layer"]
        if lay_diff < 0:
            direction_sign = -1
        else:
            direction_sign = 1
        for i_dum in range(abs(lay_diff) - 1):
            dummy_name = f"Dummy {edge} {i_dum}"
            dummy_layer = dig.nodes[v_par]["layer"] + direction_sign * (i_dum + 1)
            dig.add_node(
                dummy_name,
                layer=dummy_layer,
                in_value=edge_val,
                out_value=edge_val,
                max_value=edge_val,
            )

            if verbose:
                print(f"Adding {dummy_name}")
            if i_dum == 0:
                prev_node = edge[0]
            else:
                prev_node = f"Dummy {edge} {i_dum-1}"
            dig.add_edge(prev_node, dummy_name, **edge_dict)
            if i_dum == abs(lay_diff) - 2:
                dig.add_edge(dummy_name, edge[1], **edge_dict)
        dig.remove_edge(v_par, v_chi)


def initialize_relative_vertical_position(dig, key_name=None, verbose=False):
    """Assign/initialize relative vertical position (integer) within each layer in the simplest possible way

    Args:
        dig
        key_name: None or a string on which to sort nodes of each layer

    Returns:
        None
    """

    layers = get_layers(dig)
    if verbose:
        print(layers)

    layer_node_dict = {
        lay: get_layer_nodes(dig, lay, key_name=key_name) for lay in layers
    }
    if verbose:
        print(layer_node_dict)

    for n in dig.nodes:
        node_layer = dig.nodes[n]["layer"]
        dig.nodes[n]["vertical_position"] = layer_node_dict[node_layer].index(n)

    return layers, layer_node_dict


def edges_cross(dig: nx.DiGraph, edge_a: tuple, edge_b: tuple, verbose: bool = False):
    """Determine if two edges cross"""
    z_a = (
        dig.nodes[edge_a[0]]["vertical_position"],
        dig.nodes[edge_a[1]]["vertical_position"],
    )
    z_b = (
        dig.nodes[edge_b[0]]["vertical_position"],
        dig.nodes[edge_b[1]]["vertical_position"],
    )
    vert_pos_diff_from = (
        dig.nodes[edge_b[0]]["vertical_position"]
        - dig.nodes[edge_a[0]]["vertical_position"]
    )
    vert_pos_diff_to = (
        dig.nodes[edge_b[1]]["vertical_position"]
        - dig.nodes[edge_a[1]]["vertical_position"]
    )
    cross = (vert_pos_diff_from * vert_pos_diff_to) < 0
    if cross and verbose:
        print(f"edge_a: {edge_a} {z_a}. edge_b: {edge_b} {z_b}")
        print(f"Vertical position difference on first layer: {vert_pos_diff_from}")
        print(f"Vertical position difference on second layer: {vert_pos_diff_to}")
    return cross


def count_crossing_edges(dig: nx.DiGraph, verbose=False) -> int:
    """Count the number of crossing edge pairs in a directed graph"""
    n_crossing = 0
    layers = get_layers(dig)
    for prev_lay, next_lay in zip(layers[:-1], layers[1:]):
        # print(f'Layer {prev_lay}')
        layer_nodes = get_layer_nodes(dig, prev_lay, key_name="vertical_position")
        next_layer_nodes = get_layer_nodes(dig, next_lay, key_name="vertical_position")
        layer_edges = [
            edge
            for edge in dig.edges
            if edge[0] in layer_nodes and edge[1] in next_layer_nodes
        ]
        if verbose:
            print(f"Edges between layers {prev_lay} and {next_lay}: {layer_edges}")
        crossing_edges = [
            (edge_a, edge_b)
            for edge_a, edge_b in itertools.combinations(layer_edges, 2)
            if edges_cross(dig, edge_a, edge_b)
        ]
        if verbose:
            print(f"Crossing edge pairs: {crossing_edges}")
        n_crossing += len(crossing_edges)
    return n_crossing


def apply_barycenter_crossing_reduction_heuristic(
    dig: nx.DiGraph, from_the_left: bool = True, verbose: bool = False
):
    """Apply crossing reduction technique

    Heuristic for crossing reduction:
    changing position of nodes in layer one layer l according to the barycenter of the connected nodes in the adjacent layer (l-1 or l+1)

    Args:
        from_the_left: a boolean, True if position in a layer is changed on the basis of previous layer (left layer).
            If so, we start from the leftmost layer
    """
    #
    layers = get_layers(dig)
    if verbose:
        if from_the_left:
            print("From the left")
        else:
            print("From the right")
    if not from_the_left:
        layers = layers[::-1]

    for lay, other_lay in zip(layers[1:], layers[:-1]):
        if verbose:
            print(
                f"Changing vertical position in layer {lay} ",
                "based on connected nodes in layer {other_lay}",
            )
        layer_nodes = get_layer_nodes(dig, lay, key_name="vertical_position")
        other_layer_nodes = get_layer_nodes(
            dig, other_lay, key_name="vertical_position"
        )
        if from_the_left:
            layer_edges = [
                edge
                for edge in dig.edges
                if edge[1] in layer_nodes and edge[0] in other_layer_nodes
            ]
        else:
            layer_edges = [
                edge
                for edge in dig.edges
                if edge[0] in layer_nodes and edge[1] in other_layer_nodes
            ]
        if verbose:
            print(layer_edges)
        relative_values = []
        for node in layer_nodes:
            if from_the_left:
                connected_nodes = [edge[0] for edge in layer_edges if edge[1] == node]
            else:
                connected_nodes = [edge[1] for edge in layer_edges if edge[0] == node]
            connected_values = [
                dig.nodes[cn]["vertical_position"] for cn in connected_nodes
            ]
            relative_values.append(np.mean(connected_values))
        # add current vertical position with small weight to avoid equal values
        eps = 0.1 / len(layer_nodes)
        relative_values = np.array(relative_values) + eps * np.array(
            [dig.nodes[node]["vertical_position"] for node in layer_nodes]
        )

        if verbose:
            rel_val_dict = dict(zip(layer_nodes, relative_values))
            print(f"Relative values: {rel_val_dict}")

        # sort values, that is those that are not null
        sorted_values = sorted(relative_values)

        non_null_ids = [
            i_val for i_val, val in enumerate(relative_values) if ~np.isnan(val)
        ]
        non_null_vals = [
            val for i_val, val in enumerate(relative_values) if ~np.isnan(val)
        ]
        sorted_values = sorted(non_null_vals)
        new_values = []
        for i, val in enumerate(relative_values):
            if ~np.isnan(val):
                new_value = non_null_ids[sorted_values.index(val)]
            else:
                new_value = i
            new_values.append(new_value)
        new_val_dict = dict(zip(layer_nodes, new_values))
        if verbose:
            print(f"New relative values: {new_val_dict}")
        for i_node, node in enumerate(layer_nodes):
            dig.nodes[node]["vertical_position"] = new_values[i_node]


def sweep_barycenter_crossing_reduction(
    dig: nx.DiGraph, n_sweep_min: int = 2, n_sweep_max: int = 5, verbose: bool = False
) -> None:
    """Apply barycenter crossing reduction a number of times in each direction

    Apply it at least n_sweep_min times, then as long as the number of crossings
    is reduced up to n_sweep_max

    Args:
        dig
        n_sweep_min: the minimum number of times the crossing reduction heuristic is applied
        n_sweep_max: the maximum number of times the crossing reduction heuristic is applied

    Returns:
        None, dig is modified in place
    """
    n_cross_before = count_crossing_edges(dig)
    print(f"Edge crossing before: {n_cross_before}")
    for i_sweep in range(n_sweep_max):
        try:
            apply_barycenter_crossing_reduction_heuristic(
                dig, from_the_left=(i_sweep % 2 == 0), verbose=verbose
            )
            n_cross_after = count_crossing_edges(dig)
            print(f"Edge crossing after: {n_cross_after}")
            if n_cross_after >= n_cross_before and i_sweep >= n_sweep_min:
                break
            n_cross_before = n_cross_after
        except Exception as excep:
            print(f"Failed applying barycenter edge crossing reduction! {excep}")
            draw.plot_node_positions(dig, pos_var="vertical_position")
            apply_barycenter_crossing_reduction_heuristic(
                dig, from_the_left=(i_sweep % 2 == 0), verbose=verbose
            )


def assign_absolute_vertical_position(dig: nx.DiGraph) -> None:
    """Assign vertical position for each node based on relative vertical position

    Args:
        dig: a directed graph with the node attribute "vertical_position"

    Returns:
        None, dig is modified in place
    """

    layers = get_layers(dig)

    layer_node_dict = {
        lay: get_layer_nodes(dig, lay, key_name="vertical_position") for lay in layers
    }

    for lay in layers:
        node_y = 0
        layer_nodes = layer_node_dict[lay]
        for i_ln, node in enumerate(layer_nodes):
            if i_ln == 0:
                node_y = dig.nodes[node]["max_value"] / 2
            else:
                prev_n = layer_nodes[i_ln - 1]
                node_y += dig.nodes[prev_n]["max_value"] + dig.nodes[node]["max_value"]
            dig.nodes[node]["y"] = node_y


def process_directed_graph(
    dig: nx.DiGraph, layout_config: Optional[LayoutConfig] = None
):
    """Process directed graph to prepare Sankey diagram

    Args:
        dig: a directed graph where each node has a "max_value" attribute
        layout_method
    """

    if layout_config is None:
        layout_config = LayoutConfig()

    assign_layers(dig, method=layout_config)
    # reintroduce cycle edges
    if hasattr(dig, "cycle_edges"):
        for edge, edge_attrs in dig.cycle_edges.items():
            dig.add_edge(*edge, **edge_attrs)

    # recalculate node values after reintroducing cycle edges
    calculate_node_values(dig)
    add_dummy_nodes(dig)

    vertical_pos_method = layout_config.vertical_positioning
    initialize_relative_vertical_position(dig)
    if vertical_pos_method in ["barycenter_heuristic", "lp"]:
        n_sweep_min = layout_config.barycenter_heuristic_n_sweep_min
        n_sweep_max = layout_config.barycenter_heuristic_n_sweep_max
        sweep_barycenter_crossing_reduction(
            dig, n_sweep_min=n_sweep_min, n_sweep_max=n_sweep_max
        )

    if vertical_pos_method == "barycenter_heuristic":
        assign_absolute_vertical_position(dig)
    elif vertical_pos_method == "lp":
        optim.optimize_absolute_vertical_position_lp(dig)
    elif vertical_pos_method == "milp":
        optim.optimize_vertical_position_milp(dig)
    else:
        raise ValueError("Unknown method")



"""High-level functions for Sankey diagrams in one or few lines of code"""

from typing import Optional

import matplotlib
import networkx as nx
import pandas as pd

from . import processing
from . import draw


def sankey_from_dig(
    dig: nx.DiGraph,
    layout_config: Optional[processing.LayoutConfig] = None,
    ax: Optional[matplotlib.axes.Axes] = None,
    draw_config: Optional[draw.DrawConfig] = None,
):
    """Plot Sankey diagram from directed graph

    Args:
        dig: a directed graph with edge attribute "value"
        layout_config: layout configuration
        ax: matplotlib axes
        draw_config: drawing configuration
    """

    processing.check_digraph_before_processing(dig)
    processing.calculate_node_values(dig)
    processing.process_directed_graph(dig, layout_config)
    draw.draw_sankey(dig, ax, draw_config)


def sankey_from_df(
    flow_df: pd.DataFrame,
    layout_config: Optional[processing.LayoutConfig] = None,
    ax: Optional[matplotlib.axes.Axes] = None,
    draw_config: Optional[draw.DrawConfig] = None,
):
    """Plot Sankey diagram from dataframe

    Args:
        flow_df: a dataframe with columns "source", "target" and "value", one row per flow
        layout_config: layout configuration
        ax: matplotlib axes
        draw_config: drawing configuration
    """

    dig = processing.get_graph_from_df(flow_df)
    processing.check_digraph_before_processing(dig)
    processing.calculate_node_values(dig)
    processing.process_directed_graph(dig, layout_config)
    draw.draw_sankey(dig, ax, draw_config)
